/*  Interface to car cd-changer emulator for Sony Unilink

  18.11.2002: ad, initial implementation.

    Copyright (c)2002 by Mathias Adam, m.adam@adamis.de

	  This program is free software; you can redistribute it and/or
	  modify it under the terms of the GNU General Public License as
	  published by the Free Software Foundation; either version 2 of
	  the License, or (at your option) any later version. Read the
	  full License at http://www.gnu.org/copyleft for more details.

*/

// include files -----
//
#include <conio.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>

#include "clib/rtos.h"
#include "clib/fossil.h"

#include "suart.h"
#include "changer.h"


// reg: 0..127
bool chg_ReadRegister(unsigned char reg, unsigned char *value)
{
//  return suart_Tx91RxChar(reg + 0x80, value);
  unsigned char ch, ch2;
  for (int i = 1; i <= 30; i++) {
    if (suart_Tx91RxChar(reg + 0x80, value) &&
      suart_Tx91RxChar(reg + 0x80, &ch) && (ch == *value) &&
      suart_Tx91RxChar(reg + 0x80, &ch) && (ch == *value))        // read three times to avoid read errors
      return true;                                      // break when OK
    RTX_Sleep_Time(1);               // avoid watchdog reset when PIC doesn't answer
  }
  return false;
}

// reg: 0..127
bool chg_WriteRegister(unsigned char reg, unsigned char value)
{
  unsigned char ch;
  for (int i = 1; i <= 50; i++) {
    if (suart_Tx91RxChar(reg, &ch) && (ch == reg)) {   // send register address; retry on error
      if (suart_Tx90RxChar(value, &ch) && (ch == value)) {// send data value
        if ((reg /*> 15*/== 0x3f) || (chg_ReadRegister(reg, &ch) && (ch == value)))  // check value if register is readable; otherwise assume write OK
          return true;                                      // break when OK
      }
    }
    RTX_Sleep_Time(3);               // avoid watchdog reset when PIC doesn't answer
  }
  return false;                                         // report error after 50 fails
}

bool chg_SendDiskName(char *DiskName)
{
  int i;
  for (i = 0; (i <= 7) && (DiskName[i]); i++) {
    if (!chg_WriteRegister(0x40 + i, DiskName[i]))   ;
//      printf("write disk name char failed!\n");
//      return false;
    RTX_Sleep_Time(1);               // don't hurry; give PIC some time
//    printf("%i", i);
  }
  for (; i <= 7; i++) {
    if (!chg_WriteRegister(0x40 + i, ' '))   ;
//      printf("write disk name char failed!\n");
//      return false;
    RTX_Sleep_Time(1);
//    printf("%i", i);
  }
  return true;
}


unsigned char RecBuf[256];
int ReceivedCount = 0;
int chg_ReadDebugLog(char *l)
{
  while (fossil_peek_input(FOSSIL_EXT) >= 0) {
    RecBuf[ReceivedCount] = fossil_getbyte_wait(FOSSIL_EXT);
    if ((RecBuf[ReceivedCount] >= 0xAD) && (RecBuf[ReceivedCount] <= 0xAF)) {  // packet starts with AD/AE/AF
      int i = ReceivedCount;
      memcpy(l, RecBuf, ReceivedCount);
      RecBuf[0] = RecBuf[ReceivedCount];
      ReceivedCount = 1;
      return i;

/*
      unsigned char c = RecBuf[ReceivedCount];
      RecBuf[ReceivedCount] = 0;
      strcpy(l, RecBuf);
      ReceivedCount = 1;
      RecBuf[0] = c;
      if (strlen(l)) return true;

/*
    if (((RecBuf[ReceivedCount-1] >= 0xAD) && (RecBuf[ReceivedCount-1] <= 0xAF)) || (ReceivedCount >= 30)) {
//    if ((ReceivedCount >= 30)) {
      RecBuf[ReceivedCount] = 0;
      ReceivedCount = 0;
      strcpy(l, RecBuf);
      return true;
      */
    } else
      ReceivedCount++;
  }
  return 0;
}

