/*  Ericsson control interface

  2004-02-28: ad, initial implementation.

    Copyright (c)2004 by Mathias Adam, m.adam@adamis.de

	  This program is free software; you can redistribute it and/or
	  modify it under the terms of the GNU General Public License as
	  published by the Free Software Foundation; either version 2 of
	  the License, or (at your option) any later version. Read the
	  full License at http://www.gnu.org/copyleft for more details.

*/

// include files -----
//
#include <mem.h>
#include <malloc.h>
#include <stdio.h>
#include <math.h>
#include <conio.h>
#include <string.h>
#include <io.h>
#include <stdlib.h>
#include <dir.h>

#include "clib/rtos.h"
#include "clib/fossil.h"

#include "pcore.h"


// Eigene Datenbank hier, damit am Telefon der Verzeichnisbaum unabhngig
// vom Radio und dem Abspielvorgang durchgeblttert werden kann

int MenuLevel;           // Current directory level: 1..pc_MaxLevel

dbTable tblDirsE[pc_MaxLevel+1], tblFilesE;
DirsRec *recDirE[pc_MaxLevel+1];
FilesRec *recFileE;


#define MaxBlockSize 10

bool PhoneConnected = false;
int CurrentBlock = 0;

void fos_SendLine(char *l)
{
  for (int i=0; l[i]; i++) {
    fossil_putbyte_wait(FOSSIL_EXT, l[i]);
//    RTX_Sleep_Time(5);
  }
  fossil_putbyte_wait(FOSSIL_EXT, 13);
  fossil_putbyte_wait(FOSSIL_EXT, 10);
}

char RecLnBuf[256];
int NumReceived = 0;
bool fos_ReadLine(char *l)
{
  while (fossil_peek_input(FOSSIL_EXT) >= 0) {
    RecLnBuf[NumReceived] = fossil_getbyte_wait(FOSSIL_EXT);
    RecLnBuf[++NumReceived] = 0;
    if ((RecLnBuf[NumReceived-1] == 10) || (RecLnBuf[NumReceived-1] == 13)) {
      RecLnBuf[NumReceived-1] = 0;   // strip off newline or carriage return character
      NumReceived = 0;
      if (RecLnBuf[0]) {  // ignore empty lines
        strcpy(l, RecLnBuf);
        return true;
      }
    }
  }
  return false;
}

//char ReceiveBuffer[100];

// pcore.cpp:pc_Init() must be called before!!
void InitEricsson()
{
  PhoneConnected = false;
//  fossil_writeblock(FOSSIL_EXT, "AT*EAM=?\r\n", 10);
  fos_SendLine("AT*EAM=?");
/*  while (!fos_ReadLine(ReceiveBuffer));
  printf("\nFOSSIL: \"%s\"\n", ReceiveBuffer);
  while (!fos_ReadLine(ReceiveBuffer));
  printf("\nFOSSIL: \"%s\"\n", ReceiveBuffer);*/

  // DIRS table (pc_MaxLevel+1 instances)  (clone from pcore.cpp, so pc_Init() must be called before InitEricsson!)
  for (int iLevel = 0; iLevel <= pc_MaxLevel; iLevel++) {
    dbCloneTable(&tblDirs[0], &tblDirsE[iLevel]);
    recDirE[iLevel] = (DirsRec*)tblDirsE[iLevel].CurrentRecord;
  }
  // FILES table
  dbCloneTable(&tblFiles, &tblFilesE);
  recFileE = (FilesRec*)tblFilesE.CurrentRecord;
}

// _unsigned_ char ist hier wichtig, da sonst die <128-Abfrage vermutlich >=0 heissen msste
void utf8_to_t610(unsigned char *dest, unsigned char *src)
{
  int i = 0, j = 0;
  unsigned int utfcode;

  while (src[i]) {
    if (src[i] < 128) {
      // only exclude codes that (may) lead to errors: some codes below 32, " and \
      // TODO: there are some more codes that are different between ASCII and T610 that should be mapped/removed
      if (src[i] >= 32 && src[i] != 34 && src[i] != 92) {
        dest[j++] = src[i];
      }
    } else {
      utfcode = src[i];
      if (!src[++i]) break;  // knnte bei fehlerhaften UTF8-Strings auftreten
      utfcode += src[i] << 8;
      switch (utfcode) {
        case 0x84C3: dest[j++] = 196; break;     // 
        case 0x96C3: dest[j++] = 214; break;     // 
        case 0x9CC3: dest[j++] = 220; break;     // 
        case 0x94C3: dest[j++] = 223; break;     // 
        case 0xA0C3: dest[j++] = 224; break;     // 
        case 0xA1C3: dest[j++] = 225; break;     // 
        case 0xA4C3: dest[j++] = 228; break;     // 
        case 0xA7C3: dest[j++] = 231; break;     // 
        case 0xA8C3: dest[j++] = 232; break;     // 
        case 0xA9C3: dest[j++] = 233; break;     // 
        case 0xB1C3: dest[j++] = 241; break;     // 
        case 0xB6C3: dest[j++] = 246; break;     // 
        case 0xBCC3: dest[j++] = 252; break;     // 
      }
    }
    i++;
  }
  dest[j] = 0;
}

int DirCount, FileCount, AddCount;    // wird auch bei Menauswahl-Auswertung in DoEricsson() ausgewertet
int SelectedItem = 2;

char ATCommand[256];
char s[50], fmt[20], t610[50];
void SendMenuPart()
{
  if (!MenuLevel) {
    fos_SendLine("AT*EAM=\"Adam 301\"");   // T610 goes to standby screen
    return;
  }

  DirCount = max(recDirE[MenuLevel-1]->xChildCount - CurrentBlock*MaxBlockSize, 0);
  FileCount = 0;
  AddCount = 1;   // erster Eintrag nun entweder (mehr...) oder (Alle spielen)
  if ((MenuLevel == 1) && !CurrentBlock) AddCount++;  // erste Ebene: erster Eintrag: (Zum aktuellen Titel)
  bool MoreEnd = false;
//  if (CurrentBlock) AddCount++;  // add a (more...) item at the beginning if not the first block
  if (DirCount > MaxBlockSize) {
    DirCount = MaxBlockSize;
    AddCount++;   // add a (more...) item at the end
    MoreEnd = true;
  } else {
    FileCount = recDirE[MenuLevel-1]->xFileCount; //- DirCount                 // subtract number of dirs in this block
    if (CurrentBlock*MaxBlockSize - recDirE[MenuLevel-1]->xChildCount > 0)
      FileCount -= (CurrentBlock*MaxBlockSize - recDirE[MenuLevel-1]->xChildCount);   // and number of files in blocks before
    if (DirCount+FileCount > MaxBlockSize) {
      FileCount = MaxBlockSize - DirCount;
      AddCount++;   // add a (more...) item at the end
      MoreEnd = true;
    }
  }

  utf8_to_t610(t610, recDirE[MenuLevel-1]->Name);
     // 0: prev, 1: wait, 4: cancel  ---v      (bei reject: immer previous)   |  2:wait&wait, 8:wait&cancel
  sprintf(ATCommand, "AT*EASM=\"%.15s\",%i,%i,%i", (*t610 ? t610 : "Auswahl"), (MenuLevel > 1 ? 2 : 8),
    SelectedItem, DirCount+FileCount+AddCount);
  SelectedItem = 2;   // evtl. mal noch irgendwie ndern    - 040229
  if (CurrentBlock)
    strcat(ATCommand, ",\"(mehr...)\"");
  else if (MenuLevel > 1)
    strcat(ATCommand, ",\"(Ordner abspielen)\"");
  else {
    strcat(ATCommand, ",\"(Alle abspielen)\"");
    strcat(ATCommand, ",\"(Zum aktuellen Titel)\"");
  }

  int i, i_save;   // ist das i_save hier berhaupt ntig??   - 040229
  i_save = tblDirsE[MenuLevel].Index;
  // dynamische Lngenbegrenzung (MaxLen schliet die ,"+" bzw. ,""-Zeichen noch ein!)
  int MaxLen = (250-strlen(ATCommand)-(MoreEnd?12:0)) / (DirCount+FileCount);
  sprintf(fmt, ",\"+%%.%is\"", MaxLen - 4);
  for (i = CurrentBlock*MaxBlockSize; i < CurrentBlock*MaxBlockSize + DirCount; i++) {
    dbSetIndex(&tblDirsE[MenuLevel], recDirE[MenuLevel-1]->xFirstChild + i);
    utf8_to_t610(t610, recDirE[MenuLevel]->Name);
    sprintf(s, fmt /* ",\"+%.20s\"" */, t610);
/*        if ((strlen(ATCommand) + strlen(s)) > 230) {     // send out, then start a new AT command line
          strcat(ATCommand, ",0");
          fos_SendLine(ATCommand);
          sprintf(ATCommand, "AT*EASM=\"%s\",1,1,%i", recDir[pc_Level-1]->Name, recDir[pc_Level-1]->xChildCount);
        }*/
    strncat(ATCommand, s, 240);
/* folgender Code erzeugt jeden Meneintrag mit einem eigenen AT-Command --> sehr langsam
        sprintf(ATCommand, "AT*EASM=\"%s\",1,1,1,\"%s\"", recDir[pc_Level-1]->Name, recDir[pc_Level]->Name);
        if (i+1 < recDir[pc_Level-1]->xChildCount)
          strcat(ATCommand, ",0");
        fos_SendLine(ATCommand);
        while (!fos_ReadLine(ATCommand));
        while (!fos_ReadLine(ATCommand));*/
  }
  dbSetIndex(&tblDirsE[MenuLevel], i_save);

  i_save = tblFilesE.Index;
  sprintf(fmt, ",\"%%.%is\"", MaxLen - 3);
  int FirstFileIndex = max(CurrentBlock*MaxBlockSize - recDirE[MenuLevel-1]->xChildCount, 0);
  for (i = FirstFileIndex; i < FirstFileIndex + FileCount; i++) {
    dbSetIndex(&tblFilesE, recDirE[MenuLevel-1]->xFirstFile + i);
    utf8_to_t610(t610, recFileE->Name);
    sprintf(s, fmt /* ",\"%.20s\"" */, t610);
    strncat(ATCommand, s, 240);
  }
  dbSetIndex(&tblFilesE, i_save);

  if (MoreEnd)
    strcat(ATCommand, ",\"(mehr...)\"");
  fos_SendLine(ATCommand);
}


char ReceiveBuffer[256];
char ResultCode[6];
unsigned long LastAliveTicks;

bool DoEricsson()
{
  unsigned long ticks;
  RTX_Get_System_Ticks(&ticks);

  if (!fos_ReadLine(ReceiveBuffer)) {
    if (ticks > LastAliveTicks+60000ul) {   // reset connection after 60 seconds of inactivity
      PhoneConnected = false;
//      fos_SendLine("AT*EAM=?");
      fos_SendLine("AT*EAM=\"Adam 301\"");   // close all menus
      LastAliveTicks = ticks;
    }
    return false;
  }

// we have received something, assume that there is a phone connected now
// maybe some error-checking is needed here?
  LastAliveTicks = ticks;
  if (!PhoneConnected) {
    PhoneConnected = true;
    MenuLevel = 0;
    CurrentBlock = 0;
    fos_SendLine("AT*EAM=\"Adam 301\"");
  } else {
    strncpy(ResultCode, ReceiveBuffer, 5);
    ResultCode[5] = 0;
    if (!strcmp(ResultCode, "*EAAI")) {          // Accessory menu selected
      MenuLevel = 1;
      CurrentBlock = 0;
      SelectedItem = 2;
      SendMenuPart();
    } else if (!strcmp(ResultCode, "*EAMI")) {   // Submenu item selected
      int ItemIndex;
      sscanf(&ReceiveBuffer[7], "%i", &ItemIndex);
      if ((ItemIndex == 2) && (MenuLevel == 1) && !CurrentBlock) {  // Set menu to current song
        dbSetIndex(&tblFilesE, tblFiles.Index);
        for (int iLevel = 0; iLevel < recFile->xLevel; iLevel++)
          dbSetIndex(&tblDirsE[iLevel], tblDirs[iLevel].Index);
        MenuLevel = recFile->xLevel;
        SelectedItem = recDir[recFile->xLevel-1]->xChildCount + tblFiles.Index-recDir[recFile->xLevel-1]->xFirstFile;
        CurrentBlock = SelectedItem / MaxBlockSize;
        SelectedItem %= MaxBlockSize;
        SelectedItem += (MenuLevel == 1 && !CurrentBlock) ? 3 : 2;

//      } else if ( ((ItemIndex == 1) && ((MenuLevel > 1) || CurrentBlock)) ||         - (zum aktuellen) vor (alle abspielen)
//                  ((ItemIndex == 1) && (MenuLevel == 1) && !CurrentBlock) ) {
      } else if (ItemIndex == 1) {
        if (CurrentBlock) {
          CurrentBlock--;
          SelectedItem = MaxBlockSize + 1 + (MenuLevel == 1 && !CurrentBlock ? 1 : 0);
        } else {     // play current dir immediately
//          if (MenuLevel > 1) {  // only change current dir if (Play all) was selected
          if (tblDirs[MenuLevel-1].Index != tblDirsE[MenuLevel-1].Index) {  // restart only if changing directory
            pc_EndPlay();
            pc_ToFileIndex(recDirE[MenuLevel-1]->xFirstFile);
          }
          pc_Level = MenuLevel;
        }
      } else if ((ItemIndex == MaxBlockSize+3) || ((ItemIndex == MaxBlockSize+2) && ((MenuLevel > 1) || CurrentBlock))) {
        CurrentBlock++;
      } else if (ItemIndex == 0) {
        if (MenuLevel)
          MenuLevel--;
        if (MenuLevel) {
          SelectedItem = tblDirsE[MenuLevel].Index-recDirE[MenuLevel-1]->xFirstChild;
          CurrentBlock = SelectedItem / MaxBlockSize;
          SelectedItem %= MaxBlockSize;
          SelectedItem += (MenuLevel == 1 && !CurrentBlock) ? 3 : 2;
        }
      } else if (ItemIndex <= MaxBlockSize+1 || (ItemIndex == MaxBlockSize+2 && MenuLevel == 1 && !CurrentBlock)) {
//        ItemIndex -= (CurrentBlock?1:0);  // skip first (more...) item if available
        ItemIndex -= 1;   // now there is a (Play here) item on first block too
        if ((MenuLevel == 1) && !CurrentBlock) ItemIndex--;
        if (ItemIndex <= DirCount) {   // directory was selected
          dbSetIndex(&tblDirsE[MenuLevel], recDirE[MenuLevel-1]->xFirstChild +
            CurrentBlock*MaxBlockSize + ItemIndex - 1);
          MenuLevel++;
          CurrentBlock = 0;
        } else {    // a file was selected
          ItemIndex -= DirCount;    // ItemIndex == 1 now means first file
          pc_EndPlay();
          pc_ToFileIndex(recDirE[MenuLevel-1]->xFirstFile +
            max(CurrentBlock*MaxBlockSize - recDirE[MenuLevel-1]->xChildCount, 0) + ItemIndex - 1);
        }
      } else {
        printf("\nERICSSON: Unexpected ItemIndex = %i received!\n", ItemIndex);
        ItemIndex = 1000;   // q'n'd: used as a flag
      }
      if (ItemIndex != 1000)   // "back" > 1 second should go to standby screen --> don't redraw the menu!
        SendMenuPart();
    } else {
      printf("\nFOSSIL: \"%s\"\n", ReceiveBuffer);
      if (strcmp(ResultCode, "OK") && strcmp(ResultCode, "ERROR") &&
          strncmp(ResultCode, "AT", 2)) {   // not OK, ERROR or AT... received
        SendMenuPart();
      }
    }
  }
  return true;   // evtl. nochmal berarbeiten - berhaupt ntig?
}
