/*  mp3player main software module

    2003-09-27: ad, initial implementation.

    Copyright (c)2003 by Mathias Adam, m.adam@adamis.de

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version. Read the
    full License at http://www.gnu.org/copyleft for more details.
*/

// include files -----
//
#include <mem.h>
#include <malloc.h>
#include <stdio.h>
#include <math.h>
#include <conio.h>
#include <string.h>
#include <io.h>
#include <stdlib.h>
#include <dir.h>

#include "clib/fossil.h"
#include "clib/rtos.h"
#include "tk/checkbios.h"
#include "tk/mpbuf.h"
#include "tk/mpeg.h"
#include "tk/vs1001.h"

//#include "fat32/fatfs.cpp"
#include "db.h"
#include "pcore.h"
#include "suart.h"
#include "changer.h"
#include "ericsson.h"


// read last file ID from GNUnilink device
/*int getLastFileID()
{
  unsigned char ch_userbuf0, ch_userbuf1;
  if (!chg_ReadRegister(8, &ch_userbuf0)) return -1;
  if (!chg_ReadRegister(9, &ch_userbuf1)) return -1;
  return (ch_userbuf1 << 8) | ch_userbuf0;
}   031223 */

// read last playing position (file offset in ca. ??kbytes [10 Sekunden wre OK]) from GNUnilink device
int getLastFilePos()
{
  unsigned char ch_userbuf2, ch_userbuf3;
  if (!chg_ReadRegister(10, &ch_userbuf2)) return -1;
  if (!chg_ReadRegister(11, &ch_userbuf3)) return -1;
  return (ch_userbuf3 << 8) | ch_userbuf2;
}

// reads two bytes from GNUnilink's user memory (4 bytes, so range for Adr is 0..2)
bool chg_ReadUint(unsigned char Adr, unsigned int *Value)
{
  unsigned char c0, c1;
  if (!chg_ReadRegister(Adr+8, &c0)) return false;
  if (!chg_ReadRegister(Adr+9, &c1)) return false;
  *Value = (c1 << 8) | c0;
  return true;
}

bool chg_WriteUint(unsigned char Adr, unsigned int Value)
{
  if (!chg_WriteRegister(Adr+8, Value & 0xFF)) return false;
  if (!chg_WriteRegister(Adr+9, Value >> 8)) return false;
  return true;
}

void chg_SendTrackNum()
{
  if (recFile->xLevel == pc_Level)   // file directly within current directory: show file number
    chg_WriteRegister(0x05, recDir[pc_Level-1]->xChildCount      // first tracknumbers: subdirs; files are following
        + tblFiles.Index - recDir[pc_Level-1]->xFirstFile + 1);  // TrackNum
  else    // file in some subfolder below current directory: show number of subdir
    chg_WriteRegister(0x05, tblDirs[pc_Level].Index - recDir[pc_Level-1]->xFirstChild + 1);  // TrackNum
  // Hint: pc_Level should normally never be larger than recFile->xLevel  (031221)
}

// entspricht utf8_to_t610 aus ericsson.cpp
// _unsigned_ char ist hier wichtig, da sonst die <128-Abfrage vermutlich >=0 heissen msste
void utf8_to_unilink(unsigned char *dest, unsigned char *src)
{
  int i = 0, j = 0;
  unsigned int utfcode;

  while (src[i]) {
    if (src[i] < 128) {
      // only exclude codes that (may) lead to errors: some codes below 32, " and \
      // TODO: there are some more codes that are different between ASCII and T610 that should be mapped/removed
      if (src[i] >= 32 && src[i] != 34 && src[i] != 92) {
        dest[j++] = src[i];
      }
    } else {
      utfcode = src[i];
      if (!src[++i]) break;  // knnte bei fehlerhaften UTF8-Strings auftreten
      utfcode += src[i] << 8;
      switch (utfcode) {
        case 0x84C3: dest[j++] = (unsigned)""; break;     // 
        case 0x96C3: dest[j++] = (unsigned)""; break;     // 
        case 0x9CC3: dest[j++] = (unsigned)""; break;     // 
        case 0x94C3: dest[j++] = (unsigned)""; break;     // 
        case 0xA0C3: dest[j++] = (unsigned)""; break;     // 
        case 0xA1C3: dest[j++] = (unsigned)""; break;     // 
        case 0xA4C3: dest[j++] = (unsigned)""; break;     // 
//        case 0xA7C3: dest[j++] = 231; break;     // 
        case 0xA8C3: dest[j++] = (unsigned)""; break;     // 
        case 0xA9C3: dest[j++] = (unsigned)""; break;     // 
//        case 0xB1C3: dest[j++] = 241; break;     // 
        case 0xB6C3: dest[j++] = (unsigned)""; break;     // 
        case 0xBCC3: dest[j++] = (unsigned)""; break;     // 
      }
    }
    i++;
  }
  dest[j] = 0;
}

int main(int argc, char* argv[])
{
  bool Unilink = true;    // can be turned off via telnet ('u') to avoid timeouts when not in car
     // ad/040229: Unilink is now also automatically turned off if changer interface doesn't seem to respond
  long m_old, s_old;

  if (!CheckBIOS())
    return -1;

  bool Ericsson = true;   // Ericsson interface (disabled if Unilink logging is enabled)
  bool UnilinkLog = false;
  if ((argc >= 2) && !strcmp(argv[1], "-l")) {
    Ericsson = false;
    UnilinkLog = true;
    printf("DISABLING Ericsson interface\nENABLING Unilink logging feature\nSetting EXT baud rate to 115200!\n");
    fossil_setbaud(0, 115200, 0, 8, 1);
  }

// initialization
  // player core (database)
  if (!pc_Init("B:\\DIRS.DB", "B:\\FILES2.DB", 1, 0)) {
    printf("pc_Init() failed!\n");
    return -1;
  }
  printf("Total files in database: %i\n", tblFiles.Count);

  // VS1001 mp3 decoder chip
  VS_Init(15000000);     // clock speed 15MHz

  // autochanger (unilink) interface
  suart_Init(6, 5);      // set RX = PIO??, TX = PIO??
  if (!chg_WriteRegister(0x00, 6)) {      // DiskCount := 6
    Unilink = false;
    printf("Disabling Unilink interface!\n");
  }

  // (Sony-)Ericsson mobile phone interface
  if (Ericsson) InitEricsson();

  // others
  BUF_Init(16 * 1024, 128 * 1024L);
  //randomize();    seems to calculate randseed from seconds only - always getting the same random order
  {
    unsigned long ticks;
    RTX_Get_System_Ticks(&ticks);
    srand(ticks);
  }


// restore play position, set play mode, ...
//  int lastID = getLastFileID();
//  printf("getLastFileID()=%i\n", lastID);
  unsigned int TimeOfs;
  if (Unilink) {
    unsigned int i;
    if (chg_ReadUint(0, &i)) {
      // Userbuf0+1: upper 3 bits: pc_Level-1; lower 13: file index (0..8191)
      pc_Level = (i >> 13) + 1;
      pc_ToFileIndex(i & 0x1FFF);       // also corrects pc_Level if too big
    }
    unsigned char c;
    chg_ReadRegister(7, &c);
    TimeOfs = (unsigned int)c * 60u;
    chg_ReadRegister(6, &c);
    TimeOfs += (c & 0x0F) + ((c >> 4) * 10);
  }   // das hier rum bei gelegenheit berarbeiten  - ad/040229
  int lastPos = getLastFilePos();
  printf("getLastFilePos()=%i\n", lastPos);
//(nur zum testen/031223)  if (chg_ReadUint(2, &i)) printf("chgread(2)       %i\n", i);
  unsigned int lastPosWritten = lastPos;   // speichert zuletzt geschriebenen Wert, damit nicht zu hufig
      // neu in PIC geschrieben werden muss  - evtl. diese Funktionalitt in chg_WriteReg einbauen??  - 031001

// main loop
  unsigned char last_ch_rxdisk = 0, last_ch_rxtrack = 0, last_ch_rxplaymode = 0xFF;
  int Last_Level = 0;

  while (tblFiles.Index >= 0) {    // loop normally never ends in car use
  // update display
    // Autochanger (Unilink)
    if (Unilink) {
      unsigned char Name[100];
      if (pc_Level < recFile->xLevel)
        utf8_to_unilink(Name, recDir[pc_Level]->Name);
      else
        utf8_to_unilink(Name, recFile->Name);
      chg_SendDiskName(Name);          // track name (up to 8 chars)
      chg_SendTrackNum();
      chg_WriteRegister(0x3f, 1);                            // update display: DiskNum
    }
    // Console
    printf("Starting track %i: %s\n", tblFiles.Index, recFile->Filename);
    printf("bitrate  file_pos   time\n");

  // play file
    bool b;
    if (lastPos >= 0) {
    // seek to last playing position (only used for very first file played after reboot)
      b = pc_StartPlay(lastPos * 65536);
      lastPos = -1;
    } else
      b = pc_StartPlay(0);
    if (!b) {
      printf("Cannot open file, skipping!\n");
      pc_ToFileIndex(tblFiles.Index+1);
//      pc_SelectNextTrack();  better select next track regardless of current level as there might be no
                             // correct mp3 file in some subdirectories   - 031227
      continue;
    }
    unsigned long StartTicks, CurrentTicks;  // used to detect a VS1001 hang
    RTX_Get_System_Ticks(&StartTicks);
    while (pc_DoPlay()) {
    // display time on telnet and headunit  (code noch aufrumen  - ad/031201)
      unsigned long header;
      VS_GetMPEGHeader(&header);
      FrameInfo info;
      if (MPEG_CheckHeader(header) && MPEG_ParseHeader(header, &info)) {
        unsigned int VS_s, VS_m;
        unsigned int DecodeTime = VS_GetDecodeTime();
        VS_m = (DecodeTime + TimeOfs) / 60;
        VS_s = (DecodeTime + TimeOfs) % 60;
        printf("\r   %.3lik  %.8li  %.2u:%.2u", info.bitrate/1000, pc_GetPosition(), VS_m, VS_s);

        // show time on car radio display:
        if (Unilink) {
          if (pc_Level != Last_Level) {
            Last_Level = pc_Level;
            chg_WriteRegister(0x01, pc_Level);                     // DiskNum
            chg_WriteRegister(0x04, recDir[pc_Level-1]->xChildCount
                + recDir[pc_Level-1]->xFileCount);                 // TrackCount

            // folgender Block entspricht dem Code bei Trackbeginn (hier eingefgt 040323)
            unsigned char Name[100];
            if (pc_Level < recFile->xLevel)
              utf8_to_unilink(Name, recDir[pc_Level]->Name);
            else
              utf8_to_unilink(Name, recFile->Name);
            chg_SendDiskName(Name);          // track name (up to 8 chars)

            chg_SendTrackNum();
            chg_WriteRegister(0x3f, 1);                            // update display: DiskNum
            // store pc_Level (and file index as it's the same register)
            chg_WriteUint(0, ( ((unsigned int)(pc_Level-1)) << 13) | (((unsigned int)tblFiles.Index) & 0x1FFF));
          }

          if (VS_m > 0xFF) VS_m = 0xFF;
          if (VS_m < 0) VS_m = 0;
          if (VS_s < 0) VS_s = 0;
          if ((VS_m != m_old) || (VS_s != s_old)) {
            m_old = VS_m;
            s_old = VS_s;

            chg_WriteRegister(7, VS_m);
            chg_WriteRegister(6, ((VS_s / 10) << 4) | (VS_s % 10));

/*              chg_WriteRegister(0x3f, 3);        // update Counter (and disk name if "4" was sent before
                     // dies besser vor SendDiskName, da sonst Zeitanzeige etwas unregelmig
                     // (weil SendDiskName recht lange braucht). So wird nun bei Sekunde 4,7,10,...
                     // gescrollt                           - ad/0310031745     */

#ifndef UNILINK_SCROLL
            chg_WriteRegister(0x3f, 3);        // update Counter only
#else
            // doch besser so machen:   (0310031850)
            if (VS_s % 3 == 0) {     // every 3 seconds
              chg_WriteRegister(0x3f, 4);        // update disk name
              chg_WriteRegister(0x3f, 3);        // update Counter and start transmission to head unit
            } else
              chg_WriteRegister(0x3f, 3);        // update Counter only

            if (VS_s % 3 == 2) {     // send new disk name to GNUnilink 1 second before
              int namelen = strlen(recFile->Title);
/*                if ((namelen > 8) && (namelen <= 16)) {
                  int nameindex = ((VS_s / 3) % 2) == 0 ? 0 : namelen-8;
                  printf("nameindex=%i", nameindex);
                  chg_SendDiskName(&recFile->Title[nameindex]);
                  chg_WriteRegister(0x3f, 1);        // update Counter
                }  */
              const DisplayLength = 8;    // number of characters the display can show simultaneously
              const ScrollPeriod = 3;     // scroll every ... seconds
              const ScrollAmount = 8;     // number of characters to advance each time (<= DisplayLength)
//                if (namelen > DisplayLength) {                auch kurze Namen regelmig neu senden - ad/031003
                int nameindex = (((VS_s+1) / ScrollPeriod) % ((namelen+ScrollAmount-1) / ScrollAmount)) * ScrollAmount;
                chg_SendDiskName(&recFile->Title[nameindex]);
//nein,s.o. (0310031850)                  chg_WriteRegister(0x3f, 4);        // update Counter
//                }
            }
#endif
          }
        }
      }

      if (Unilink) {
      // save current playing position regularly
        if ((pc_GetPosition() >> 16) != lastPosWritten) {
          lastPosWritten = pc_GetPosition() >> 16;
          chg_WriteRegister(10, (lastPosWritten & 0xFF));       // write register 10 (UserBuf2)
          chg_WriteRegister(11, lastPosWritten >> 8);           // write register 11 (UserBuf3)
        }

      // poll GNUnilink for control commands
        unsigned char ch_rxdisk, ch_rxtrack, ch_rxtrackRel, ch_rxplaymode, ch;
        if (chg_ReadRegister(13, &ch_rxtrack) && ch_rxtrack) {   // TrackNum received
          chg_WriteRegister(13, 0);                // clear register
          printf("\nUNILINK: New track number = %hi", ch_rxtrack);
          // Becker sends disk number follwoing track=1 when a new MD was selected:
          ch_rxdisk = 0;
          if ((ch_rxtrack == 1) && chg_ReadRegister(12, &ch_rxdisk) && (ch_rxdisk)) {    // DiskNum received
            chg_WriteRegister(12, 0);                // clear register
          }
          if (ch_rxdisk && (ch_rxdisk != pc_Level)) {    // DiskNum has changed: ignore track number
            if (ch_rxdisk <= 4) {  // Disk 1..4: Select level
              if (ch_rxdisk > recFile->xLevel) ch_rxdisk = recFile->xLevel;
              pc_Level = ch_rxdisk;
            }
          } else {                                 // evaluate track value:
            if (ch_rxtrack <= recDir[pc_Level-1]->xChildCount) {   // selected item is a subdirectory
              pc_EndPlay();
              pc_SelectSubdir(ch_rxtrack-1);
              pc_ToFileIndex(recDir[pc_Level]->xFirstFile);
            } else {   // a file was selected
              pc_EndPlay();
              pc_ToFileIndex(recDir[pc_Level-1]->xFirstFile - 1 + ch_rxtrack - recDir[pc_Level-1]->xChildCount);
            }
          }
        }
        if (chg_ReadRegister(15, &ch_rxplaymode)) {  // PlayMode received
/*          if (ch_rxplaymode & 0x01) {                // RPT mode selected
            PlayMode = REPEATGROUP;		  // Wiederhole Verzeichnis
          } else*/ if (ch_rxplaymode & 0x02) {         // RP mode selected
            pc_Playmode = RANDOM;    	      	  // Zufall (ber alle Dateien)
          } else {                						  // normal mode
            pc_Playmode = PLAY;              // alle Titel nacheinander
          }
        }
        if (chg_ReadRegister(14, &ch_rxtrackRel) && ch_rxtrackRel) {   // TrackRel received
          chg_WriteRegister(14, 0);     // clear register 14 (TrackRel)
          pc_Playmode = RANDOM;                        // Becker benutzt dies nur im RP-Modus
                   // Achtung: Playmode wird bei nchster nderung am Playmode-Register im
                   // PIC wieder zurckgesetzt! Besser: PIC sollte sein Playmode-Bit
                   // entsprechend setzen, wenn er einen TrackRel-Befehl empfngt; oder
                   // noch besser: PIC stabilisieren, damit er alles vom Radio korrekt
                   // empfngt                                             - 031001
          pc_EndPlay();
          pc_SelectNextTrack();
          printf("\nnext track was selected by TrackRel\n");
          break;
        }
      }

    // check for console input
      if (kbhit()) {
        int i, i_save;  // case 'd', 'f'
        switch (getch()) {
          case 'u':
            Unilink = !Unilink;
            printf("\nUnilink communication now %sabled.\n", Unilink ? "en" : "dis");
            break;

          case 'r':
            pc_Playmode = ((pc_Playmode == PLAY) ? RANDOM : PLAY);
            printf("\nNew playmode: %s\n", (pc_Playmode == PLAY) ? "PLAY" : "RANDOM");
            break;

          case 'd':  // list subdirectories in current dir
            printf("\nSubdirectories of %s:\n", recDir[pc_Level-1]->Name);
            i_save = tblDirs[pc_Level].Index;
            for (i = 0; i < recDir[pc_Level-1]->xChildCount; i++) {
              dbSetIndex(&tblDirs[pc_Level], recDir[pc_Level-1]->xFirstChild + i);
              printf("%4i. %s\n", i+1, recDir[pc_Level]->Name);
            }
            dbSetIndex(&tblDirs[pc_Level], i_save);
            printf("Total: %i\n", recDir[pc_Level-1]->xChildCount);
            break;

          case 'f':  // list files in current dir
            printf("\nTracks in %s:\n", recDir[pc_Level-1]->Name);
            i_save = tblFiles.Index;
            for (i = 0; i < recDir[pc_Level-1]->xFileCount; i++) {
              dbSetIndex(&tblFiles, recDir[pc_Level-1]->xFirstFile + i);
              printf("%4i. %s\n", i+1, recFile->Name);
            }
            dbSetIndex(&tblFiles, i_save);
            printf("Total: %i\n", recDir[pc_Level-1]->xFileCount);
            break;

          case 'l':
            printf("\nCurrent Level: %i / Enter new level (1..%i): ", pc_Level, recFile->xLevel);
            int iNewTrack;
            scanf("%i", &iNewTrack); puts("");
            if ((iNewTrack >= 1) && (iNewTrack <= recFile->xLevel))
              pc_Level = iNewTrack;
            break;

          case 'n':
            pc_EndPlay();
            pc_SelectNextTrack();
            break;

          case 's':
            if (recDir[pc_Level-1]->xChildCount <= 0) {
              printf("\nNo subdirectories! (Current Level = %i)\n", pc_Level);
              break;
            }
            printf("\nSelect subdirectory (1..%i): ", recDir[pc_Level-1]->xChildCount);
//            int iNewTrack;
            scanf("%i", &iNewTrack);
            // neuer code (nach nderung xFirstFile: wird nun auch bei leeren Parent-Directories belegt)
            if (pc_SelectSubdir(iNewTrack-1)) {
              pc_EndPlay();
              pc_ToFileIndex(recDir[pc_Level]->xFirstFile);
            }
#ifdef alskjdflksj     //alter code
            if (pc_SelectSubdir(iNewTrack-1)) {
              pc_Level++;
              printf("\nEntering: %s\\", recDir[pc_Level-1]->Name);
              pc_EndPlay();
              while (recDir[pc_Level-1]->xFileCount <= 0) {  // assumes there are no empty subdirectories!
                pc_SelectSubdir(0);
                pc_Level++;
                printf("%s\\", recDir[pc_Level-1]->Name);
              }
              printf(" (%i files)\n", recDir[pc_Level-1]->xFileCount);
              pc_ToFileIndex(recDir[pc_Level-1]->xFirstFile);
            }
#endif
            break;

          case 't':
            if (recDir[pc_Level-1]->xFileCount <= 0) {
              printf("\nNo files in this directory; use 's' to select a subdir! (Current Level = %i)\n", pc_Level);
              break;
            }
            printf("\nEnter track number (1..%i): ", recDir[pc_Level-1]->xFileCount);
//            int iNewTrack;
            scanf("%i", &iNewTrack);
            pc_EndPlay();
            pc_ToFileIndex(recDir[pc_Level-1]->xFirstFile - 1 + iNewTrack);
            break;

          case 'z':
            printf("\nEnter new track number (global / 0..%i): ", tblFiles.Count-1);
//            int iNewTrack;
            scanf("%i", &iNewTrack);
            pc_EndPlay();
            pc_ToFileIndex(iNewTrack);
            break;

          default:
            tblFiles.Index = -1;   // stop playback
            pc_EndPlay();
        }
      }

    // Ericsson interface: check for control commands
      if (Ericsson) DoEricsson();

    // Unilink debug feature: show log messages from GNUnilink (EXT_BAUD=115200, Jumper set to PIC)
      unsigned char LogMessage[100];
      int msglen;
      LogMessage[0] = 0;   // only a flag if to print the last \n
      while (UnilinkLog && (msglen = chg_ReadDebugLog(LogMessage))) {
        printf("\nUNILINK: ");
        for (int i = 0; i < msglen; i++)
          printf("%02hX ", LogMessage[i]);
      }
      if (LogMessage[0])
        printf("\n");

    // Pause if GNUnilink says that we're deselected
      if (Unilink) {
        unsigned char ch_rxplaymode, ch;
        if (chg_ReadRegister(15, &ch_rxplaymode) && !(ch_rxplaymode & 0x80)) {
          printf("\nUnilink says we're deselected, pausing...");
          // VS may keep playing for several seconds. Just leave it this way as it might be
          // rather complicated (stopping and restarting DMA?) to fix this    - ad/040331
          int i = 0;
          while (chg_ReadRegister(15, &ch_rxplaymode) && !(ch_rxplaymode & 0x80)) {
            // evtl. hier Aufrumarbeiten vornehmen / Ericsson behandeln?
            RTX_Sleep_Time(100);
            if (i++ < 120*10) suart_TxChar_91(0xFF);
          }
          printf("continuing.\n");
        }
      }

    // be nice to other apps
      RTX_Sleep_Time(5);
    }
    TimeOfs = 0;  // only used for the first track (offset according to lastPos)
    puts("");  // newline

    if (tblFiles.Index >= 0)       // save track ID (only if != -1) and pc_Level (as it's the same register)
      chg_WriteUint(0, ( ((unsigned int)(pc_Level-1)) << 13) | (((unsigned int)tblFiles.Index) & 0x1FFF));

    if (kbhit()) {           // allow to quit even if file cannot be loaded
      switch (getch()) {
        case 'u','r','l','n','s','t','z':
          break;

        default:
          tblFiles.Index = -1;   // stop playback
      }
    }
  }


// clean up
  pc_Done();
  BUF_Done();
  VS_Done();

  puts("MP3.EXE finished.\n");
}
