/*  player core

  2003-12-20: ad, initial implementation.

    Copyright (c)2003 by Mathias Adam, m.adam@adamis.de

	  This program is free software; you can redistribute it and/or
	  modify it under the terms of the GNU General Public License as
	  published by the Free Software Foundation; either version 2 of
	  the License, or (at your option) any later version. Read the
	  full License at http://www.gnu.org/copyleft for more details.

*/

// include files -----
//
#include <mem.h>
#include <malloc.h>
#include <stdio.h>
#include <math.h>
#include <conio.h>
#include <string.h>
#include <io.h>
#include <stdlib.h>
#include <dir.h>

#include "tk/mpbuf.h"
#include "tk/mpeg.h"
#include "tk/vs1001.h"

#include "pcore.h"
#include "fat32/fatfs.cpp"


int pc_Playmode = PLAY;

int pc_Level;           // Current directory level: 1..pc_MaxLevel

dbTable tblDirs[pc_MaxLevel+1], tblFiles;
DirsRec *recDir[pc_MaxLevel+1];
FilesRec *recFile;




int mp3file;       // file handle
bool IsPlaying = false;
long fpos = 0;

/*struct FilesRec {       // all strings are null-terminated!
  char Artist[65];
  char Album[65];
  char Title[64];
  unsigned __int8 TrackNum;
  char Comment[65];
  char Year[5];
};

struct DirectoryRec {
  int Parent;           // signed 16bit: -1 = root dir is parent
  char Name[100];
}

struct FileRec {
  int Parent;           // signed 16bit: -1 = root dir is parent
  char Name[100];
} */



// NewLevel: 0..pc_MaxLevel (range isn't checked here!)
// FileIndex: 0..tblFiles.Count-1
bool pc_Init(char *fnDirs, char *fnFiles, int NewLevel, int FileIndex)
{
  // DIRS table (pc_MaxLevel+1 instances)
  if (dbOpenTable(fnDirs, &tblDirs[0], sizeof(DirsRec))) {
    printf("Cannot open %s.\n", fnDirs);
    return false;
  }
  recDir[0] = (DirsRec*)tblDirs[0].CurrentRecord;
  for (int iLevel = 1; iLevel <= pc_MaxLevel; iLevel++) {
    dbCloneTable(&tblDirs[0], &tblDirs[iLevel]);
    recDir[iLevel] = (DirsRec*)tblDirs[iLevel].CurrentRecord;
  }
  // FILES table
  if (dbOpenTable(fnFiles, &tblFiles, sizeof(FilesRec))) {
    dbCloseTable(&tblDirs[0]);
    // hier noch die Clones freigeben    - 031222
    printf("Cannot open %s.\n", fnFiles);
    return false;
  }
  recFile = (FilesRec*)tblFiles.CurrentRecord;

  pc_Level = NewLevel;
  return pc_ToFileIndex(FileIndex);

//  return true;
}

void pc_Done()
{
  dbCloseTable(&tblDirs[0]);
  // hier noch die Clones freigeben    - 031222
  dbCloseTable(&tblFiles);
}


bool pc_ToFileIndex(int NewIndex)
{
  if (dbSetIndex(&tblFiles, NewIndex)) return false;
  dbSetIndex(&tblDirs[recFile->xLevel-1], recFile->Parent);
  for (int iLevel = recFile->xLevel-2; iLevel > 0; iLevel--) {
    dbSetIndex(&tblDirs[iLevel], recDir[iLevel+1]->Parent);
  }
  if (pc_Level > recFile->xLevel)
    pc_Level = recFile->xLevel - 1;
  return true;
}


bool pc_SelectSubdir(int ChildIndex)
{
  if ((ChildIndex < 0) || (ChildIndex >= recDir[pc_Level-1]->xChildCount)) return false;
  dbSetIndex(&tblDirs[pc_Level], recDir[pc_Level-1]->xFirstChild + ChildIndex);
  return true;
}


// selects first file under current directory, starting at pc_Level (searches subdirectories too)
// Assumes that there are no empty subdirectories!
bool pc_SelectFirstFile()
{
  int LevelSave = pc_Level;
  while (recDir[pc_Level-1]->xFileCount <= 0) {  // assumes there are no empty subdirectories!
    pc_SelectSubdir(0);
    pc_Level++;
  }
  pc_ToFileIndex(recDir[pc_Level-1]->xFirstFile);
  pc_Level = LevelSave;
}


bool pc_StartPlay(long StartPos)
{
  char mp3filename[260];
  strcpy(mp3filename, "C:\\");
  strcat(mp3filename, recFile->Filename);
#ifdef alsljfasdf
    for (iLevel = 1; iLevel < recFile->xLevel; iLevel++) {
      strcat(mp3filename, recDir[iLevel-1]->Name);
      strcat(mp3filename, "\\");
    }
    if (recFile->TrackNum) {
      char chtracknum[5];
      sprintf(chtracknum, "%.2hu ", recFile->TrackNum);
      strcat(mp3filename, chtracknum);
    }
    strcat(mp3filename, recFile->Title);
    strcat(mp3filename, ".mp3");
#endif
  if ((mp3file = fs_open(mp3filename, FD32_OPEN_EXIST_OPEN | FD32_OPEN_ACCESS_READ, NULL, 0x27, 1)) < 0) return false;
  VS_Reset();
  BUF_Play();
  IsPlaying = true;
  fpos = StartPos;
  if (fpos > 0) fs_lseek(mp3file, &fpos, FD32_SEEK_SET);

  return true;
}

// Should be called regularly (multiple times a second) during playback.
// Reads mp3 data from file into the buffer to be sent to VS1001.
// Returns true as long as the song isn't finished; otherwise calls
// pc_EndPlay() and pc_SelectNextTrack() and returns false.
bool pc_DoPlay()
{
  if (!IsPlaying) return false;

// Read data from mp3 file
  unsigned long len;
  long len1, len2;
  char *buf1, *buf2;

  BUF_Lock(BUF_FROMWRITEPOS, min(16384L, BUF_GetFreeBytes()), &buf1, &len1, &buf2, &len2);
  len = len1;
  if (fs_read(mp3file, buf1, &len) < 0)   // re-sets len to actual number of bytes read
    puts("fs_read 1 failed!");
  if (buf2 && (len == len1)) {    // 2nd buffer available and eof not yet reached
    len = len2;
    if (fs_read(mp3file, buf2, &len) < 0)   // re-sets len to actual number of bytes read
      puts("fs_read 2 failed!");
    len += len1;
  }
  fpos += len;
  BUF_Unlock(len);

#ifdef alskdjf
      // check for decodetime showing unexpected values (more than +-5 seconds)
        RTX_Get_System_Ticks(&CurrentTicks);
        unsigned int DecodeTime = VS_GetDecodeTime();   // was within display code (below) before   ad/031201
        if ( ((unsigned long)(DecodeTime+5) < (CurrentTicks-StartTicks) / 1000) ||     // decodetime too small by 5 seconds
         ((unsigned long)(DecodeTime) > ((CurrentTicks-StartTicks) / 1000 +5))) {     // or too big by 5 seconds (presumably VS has some trouble too)
          printf("\nVS1001 stalled, skipping Song...\n");
          BUF_Done();
          VS_Done();
//          BUF_Clear();
          VS_Init(15000000);     // clock speed 15MHz
          BUF_Init(16 * 1024, 128 * 1024L);
          if (dbSetIndex(&tblFiles, tblFiles.Index+1))     // just skip to the next song
            dbSetIndex(&tblFiles, 0);                      // select track 0 after last track finished
          IsPlaying = false;
          break;
        }
#endif

  // restore settings (especially clock and decode time) after unwanted VS-resets
  VS_RecoverSettings();

  if ((len == len1+len2) || (BUF_GetState() != BUF_STALLED)) {
    return true;
  } else {
    pc_EndPlay();
    pc_SelectNextTrack();
    return false;
  }
}

// cleanup after playing a file
void pc_EndPlay()
{
// cleanup
  if (IsPlaying) {
    IsPlaying = false;
    BUF_Clear();
    fs_close(mp3file);
  }
}

// determines next file to play according to current playmode
void pc_SelectNextTrack()
{
  int DirLevel = pc_Level-1;
  if ((pc_Level == 1) && (tblFiles.Index >= recDir[0]->xTotalFiles))  // current file is a symlink and level == 1
    DirLevel++;  // act as if the subdirectory was selected
  switch (pc_Playmode) {
    case PLAY:
      if (tblFiles.Index < (recDir[DirLevel]->xFirstFile + recDir[DirLevel]->xTotalFiles - 1))
        pc_ToFileIndex(tblFiles.Index+1);
      else
        pc_ToFileIndex(recDir[DirLevel]->xFirstFile);
      break;

    case RANDOM:
      pc_ToFileIndex(recDir[DirLevel]->xFirstFile + random(recDir[DirLevel]->xTotalFiles));
      break;
  }
}


long pc_GetPosition()
{
  return fpos - BUF_GetCount();
}

